/**
 * GCodeReader.java
 * 
 * Class to read and process G-code file
 * @author Chris Meighan
 * 
 * This file is part of G-code for RepRap.
 * 
 * G-code for RepRap is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * G-code for RepRap is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with G-code for RepRap.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.chrismeighan.repgcode;

import java.io.*;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.HashMap;

public class GCodeReader {
	
	private int pointer = 0;
	
	private int filePosition;
	
	private GCodeLine currentLine;
	
	private ArrayList<GCodeLine> GCode;
	
	// Constructor
	public GCodeReader() {
		GCode = new ArrayList<GCodeLine>();
	}
	
	/**
	 * Load G-code file
	 * @param inputFile
	 * @throws Exception
	 */
	public void load(String inputFile) throws Exception {
		
		String err = "";
		
		try {
			BufferedReader inputStream = new BufferedReader(new FileReader( inputFile ));
			pointer = 0;
			GCode.clear();
			String line;
			int pos = 0;
            while ((line = inputStream.readLine()) != null) {
            	GCode.add(new GCodeLine(line,pos,line.length()));
            	pos += line.length() + 1;
            }
		} catch (FileNotFoundException e) {
			err = "File '" + inputFile + "' not found";
			throw new Exception(err);
		} catch (IOException e) {
			err = "IO Error";
			throw new Exception(err);
		}

	}
	
	/**
	 * Set file position pointer
	 */
	public void setPointer(int pointer) {
		this.pointer = pointer;
	}

	/**
	 * Read next line from G-code file
	 * @return list of word blocks
	 */
	public GCodeWordBlockList readLine() {
		
		// Check not at/past end of file
		if ( pointer >= GCode.size() ) return null;
		String line;

		// Get next line and increment pointer, skipping any blank lines
		while ( ( currentLine = GCode.get(pointer++)).getText() == "" && pointer < GCode.size()) {}

		filePosition = currentLine.getStartPos();
		
		line = currentLine.getText();
		
		// If no more non-blank lines, return
		if ( line == "" ) return null;
		
		return new GCodeWordBlockList(line);
	}
	
	/**
	 * Get current position in file
	 * @return
	 */
	public int getFilePosition() { return filePosition; }
	
	/**
	 * Return current line as text
	 * @return
	 */
	public String getCurrentLineText() { return currentLine.getText(); }
	
	/**
	 * Return the current G-code file as a string
	 * @return
	 */
	public String getGCodeAsString() {
		String GCodeString = "";
		int i;
		for (i = 0; i < GCode.size(); i++) {
			GCodeString = GCodeString.concat(GCode.get(i).getText()) + "\n";
		}
		return GCodeString;
	}
	
	/**
	 * Line of G-code
	 * @author chris
	 *
	 */
	public class GCodeLine {
		private String text;
		private int startPos;
		private int endPos;
		
		public GCodeLine( String text, int startPos, int endPos) {
			this.text = text;
			this.startPos = startPos;
			this.endPos = endPos;
		}
		
		public String getText() {
			return text;
		}
		
		public int getStartPos() {
			return startPos;
		}
		
		public int getEndPos() {
			return endPos;
		}
	}
	
	/**
	 * Block of G-code words
	 * @author chris
	 *
	 */
	public class GCodeWordBlock {
		HashMap<Character,Double> wordMap;
		
		/**
		 * Create a GCodeWordBlock from a string containing G-code
		 * @param wordBlock
		 */
		public GCodeWordBlock(String wordBlock) {
			wordMap = new HashMap<Character,Double>();
			StringTokenizer st = new StringTokenizer(wordBlock);
			String word;
			Character code;
			Double value;
			while (st.hasMoreTokens()) {
				word = st.nextToken();
				code = word.charAt(0);
				value = Double.valueOf(word.substring(1));
				wordMap.put(code, value);
			}
		}
		
		public boolean containsKey(char key) {
			return wordMap.containsKey(key);
		}
		
		public Double get(char key) {
			return wordMap.get(key);
		}
	}
	
	/**
	 * List of blocks of G-code words
	 * @author chris
	 *
	 */
	public class GCodeWordBlockList {
		ArrayList<GCodeWordBlock> words;
		
		/**
		 * Create list of blocks of G-code words,
		 * keeping parameters (e.g. X,Y,Z,R,P) together with instruction word
		 * @return
		 */
		public GCodeWordBlockList(String line) {
			words = new ArrayList<GCodeWordBlock>();
			StringTokenizer st = new StringTokenizer(line);
			String block = "", code;
			while (st.hasMoreTokens()) {
				code = st.nextToken().toUpperCase();
				char type = code.charAt(0);
				if ( type == 'N' || type == 'G' || type == 'M' || type == 'T' || type == 'S' || type == 'F' ) {
					if (block.length() > 0) words.add(new GCodeWordBlock(block));
					block = "";
				}
				if (block.length() > 0) block = block.concat(" ");
				block = block.concat(code);
			}
			if (block.length() > 0) words.add(new GCodeWordBlock(block));
		}
		
		/**
		 * Add a block of words
		 * @param word
		 */
		public void add(GCodeWordBlock word) {
			words.add(word);
		}
		
		/**
		 * Return size
		 * @return
		 */
		public int size() {
			return words.size();
		}
		
		/**
		 * Get block of words at index
		 * @param index
		 * @return
		 */
		public GCodeWordBlock get(int index) {
			return words.get(index);
		}
	}
}