/**
 * Main.java
 * 
 * Main class for G-code for RepRap
 * @author Chris Meighan
 * 
 * This file is part of G-code for RepRap.
 * 
 * G-code for RepRap is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * G-code for RepRap is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with G-code for RepRap.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.chrismeighan.repgcode;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Dimension;
import java.awt.Color;
import java.io.IOException;
import java.io.*;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.WindowConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.JTextPane;
import javax.swing.JTextField;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileFilter;
import javax.swing.JScrollPane;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.text.*;
import java.awt.event.KeyEvent;

import java.util.ArrayList;

import org.reprap.Preferences;
import org.reprap.comms.Communicator;
import org.reprap.comms.snap.SNAPAddress;
import org.reprap.comms.snap.SNAPCommunicator;
import org.reprap.devices.GenericStepperMotor;
import org.reprap.gui.Utility;

public class Main extends javax.swing.JFrame implements ChangeListener,
		ActionListener {

	private final int localNodeNumber = 0;
	
	private String version = "0.1";

	// GUI elements
	private JFrame frame;
	private JMenuBar menuBar;
	private JMenu fileMenu, helpMenu;
	private JMenuItem quitMenuItem, aboutMenuItem;
	private JButton fileOpenButton;
	private JFileChooser fc;
	private JTextField filePath;
	private JTextPane gcodeTextPane;
	private StyledDocument gcodePaneStyledDocument;
	private Style gcodePaneDefaultStyle, gcodePaneProcessedLineStyle,
			gcodePaneCurrentLineStyle;
	private JTextField statusField;
	private JButton runButton;
	private JButton emergencyStopButton;
	private JButton resetButton;
	private JButton setupButton;
	private JButton decX;
	private JButton incX;
	private JButton decY;
	private JButton incY;
	private JButton decZ;
	private JButton incZ;
	private JButton homeX;
	private JButton homeY;
	private JButton homeZ;
	private JButton homeAll;
	private JButton setDatum;
	private JLabel absPositionIndicator;
	private JLabel relPositionIndicator;
	private JTextField gcodeInputField;
	private JButton gcodeInputOK;

	// Class to read G-code files
	private GCodeReader gcodeReader;

	// Absolute position and datum offsets
	private double absX = 0, absY = 0, absZ = 0;
	private double datumXoff = 0, datumYoff = 0, datumZoff = 0;

	// Values read from properties file
	private double Xinc, Yinc, Zinc;
	private int stepsPerUnitX, stepsPerUnitY, stepsPerUnitZ;
	private int maxSpeedX, maxSpeedY, maxSpeedZ;
	private double defaultFeed;
	private double curveSection;
	private double distancePerRevolution;
	private String portName;
	private int XAxisAddress, YAxisAddress, ZAxisAddress;

	// SNAP interface
	Communicator communicator;
	private GenericStepperMotor motorX, motorY, motorZ;

	// Communicate with reprap - set false to run headless
	private boolean wantComms = true;

	// Flag to indicate if comms set up OK
	private boolean commsOK = false;

	// Flag used to tell processing thread to stop
	private boolean programRunning = false;

	// Preferences
	private GCodeRepRapPreferences p;

	public static void main(String[] args) throws Exception {
		Thread.currentThread().setName("G-code for RepRap");
		JFrame frame = new JFrame();

		try {
			Main inst = new Main(frame);
			inst.setVisible(true);
		} catch (Exception e) {
			JOptionPane.showMessageDialog(null, e.getMessage());
			return;
		}
	}

	public Main(JFrame frame) throws Exception {
		this.frame = frame;

		p = new GCodeRepRapPreferences();

		loadPreferences();

		gcodeReader = new GCodeReader();

		initGUI();
		Utility.centerWindowOnScreen(this);
	}

	/**
	 * Release all resources
	 */
	public void dispose() {
		super.dispose();
		if (motorX != null)
			motorX.dispose();
		if (motorY != null)
			motorY.dispose();
		if (motorZ != null)
			motorZ.dispose();
		if (communicator != null)
			communicator.dispose();
	}

	/**
	 * Initialise comms
	 */
	private void initialiseComms() throws Exception {
		String err = "";
		if (wantComms) {
			commsOK = false;
			try {
				SNAPAddress address = new SNAPAddress(localNodeNumber);
				communicator = new SNAPCommunicator(portName, address);
			} catch (gnu.io.NoSuchPortException e) {
				err = "There was an error opening " + portName + ".\n\n";
				err += "Check to make sure that is the right path.\n";
				err += "Check that you have your serial connector plugged in.";
				commsOK = false;
			} catch (gnu.io.PortInUseException e) {
				err = "The " + portName
						+ " port is already in use by another program.";
				commsOK = false;
			}
			if (err.length() > 0)
				JOptionPane.showMessageDialog(frame, err);
			else
				commsOK = true;
		}
	}

	/**
	 * Initialise motors
	 * 
	 * @throws IOException
	 */
	private void initialiseMotors() throws IOException {
		if (wantComms) {
			motorX = new GenericStepperMotor(communicator, new SNAPAddress(
					XAxisAddress), Preferences.getGlobalPreferences(), 1);
			motorY = new GenericStepperMotor(communicator, new SNAPAddress(
					YAxisAddress), Preferences.getGlobalPreferences(), 2);
			motorZ = new GenericStepperMotor(communicator, new SNAPAddress(
					ZAxisAddress), Preferences.getGlobalPreferences(), 3);
		}
	}

	/**
	 * Initialise GUI
	 * 
	 * @throws Exception
	 */
	private void initGUI() throws Exception {
		setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
		setTitle("G-code for RepRap");
		setResizable(false);
		
		// Menus
		menuBar = new JMenuBar();
		fileMenu = new JMenu("File");
		fileMenu.setMnemonic(KeyEvent.VK_F);
		helpMenu = new JMenu("Help");
		helpMenu.setMnemonic(KeyEvent.VK_H);
		menuBar.add(fileMenu);
		menuBar.add(helpMenu);
		quitMenuItem = new JMenuItem("Quit",KeyEvent.VK_Q);
		quitMenuItem.addActionListener(this);
		fileMenu.add(quitMenuItem);
		aboutMenuItem = new JMenuItem("About",KeyEvent.VK_A);
		aboutMenuItem.addActionListener(this);
		helpMenu.add(aboutMenuItem);
		setJMenuBar(menuBar);
		
		JPanel panel = new JPanel(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();

		c.anchor = GridBagConstraints.FIRST_LINE_START;

		// File open
		c.gridwidth = 3;
		JPanel fileOpenPanel = new JPanel(new FlowLayout());
		fileOpenButton = new JButton("Open..");
		fileOpenButton.addActionListener(this);
		fileOpenPanel.add(fileOpenButton, c);
		filePath = new JTextField(40);
		filePath.setEditable(false);
		fileOpenPanel.add(filePath, c);
		panel.add(fileOpenPanel, c);

		// Jog controls
		c.gridy = 1;
		c.gridwidth = 1;
		JPanel jogPanel = new JPanel(new GridBagLayout());
		GridBagConstraints c1 = new GridBagConstraints();
		incZ = new JButton("+Z");
		incZ.addActionListener(this);
		jogPanel.add(incZ, c1);
		c1.gridx = 2;
		incY = new JButton("+Y");
		incY.addActionListener(this);
		jogPanel.add(incY, c1);
		c1.gridx = 1;
		c1.gridy = 1;
		decX = new JButton("-X");
		decX.addActionListener(this);
		jogPanel.add(decX, c1);
		c1.gridx = 3;
		incX = new JButton("+X");
		incX.addActionListener(this);
		jogPanel.add(incX, c1);
		c1.gridx = 0;
		c1.gridy = 2;
		decZ = new JButton("-Z");
		decZ.addActionListener(this);
		jogPanel.add(decZ, c1);
		c1.gridx = 2;
		decY = new JButton("-Y");
		decY.addActionListener(this);
		jogPanel.add(decY, c1);
		panel.add(jogPanel, c);

		// Home controls
		c.gridx = 1;
		c.gridheight = 2;
		JPanel homePanel = new JPanel(new GridBagLayout());
		GridBagConstraints c2 = new GridBagConstraints();
		homeX = new JButton("Home X");
		homeX.addActionListener(this);
		homePanel.add(homeX, c2);
		c2.gridy = 1;
		homeY = new JButton("Home Y");
		homeY.addActionListener(this);
		homePanel.add(homeY, c2);
		c2.gridy = 2;
		homeZ = new JButton("Home Z");
		homeZ.addActionListener(this);
		homePanel.add(homeZ, c2);
		c2.gridy = 3;
		homeAll = new JButton("Home all");
		homeAll.addActionListener(this);
		homePanel.add(homeAll, c2);
		c2.gridy = 4;
		setDatum = new JButton("Set datum");
		setDatum.addActionListener(this);
		homePanel.add(setDatum, c2);
		panel.add(homePanel, c);

		// G-code pane
		c.gridx = 2;
		c.gridheight = 3;
		gcodeTextPane = new JTextPane();
		gcodeTextPane.setEditable(false);
		gcodeTextPane.setMinimumSize(new Dimension(300, 200));
		JScrollPane editorScrollPane = new JScrollPane(gcodeTextPane);
		editorScrollPane
				.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
		editorScrollPane.setPreferredSize(new Dimension(300, 200));
		editorScrollPane.setMinimumSize(new Dimension(10, 10));
		panel.add(editorScrollPane, c);
		//gcodeReader
		//		.load("D:\\work\\projects\\repstrap\\cad\\good_morning_both.nc");
		gcodePaneStyledDocument = gcodeTextPane.getStyledDocument();
		Style def = StyleContext.getDefaultStyleContext().getStyle(
				StyleContext.DEFAULT_STYLE);
		gcodePaneDefaultStyle = gcodePaneStyledDocument
				.addStyle("regular", def);
		StyleConstants.setFontFamily(gcodePaneDefaultStyle, "Courier");
		StyleConstants.setFontSize(gcodePaneDefaultStyle, 11);
		StyleConstants.setForeground(gcodePaneDefaultStyle, Color.black);
		StyleConstants.setItalic(gcodePaneDefaultStyle, false);
		gcodePaneProcessedLineStyle = gcodePaneStyledDocument.addStyle(
				"processed", gcodePaneDefaultStyle);
		StyleConstants.setItalic(gcodePaneProcessedLineStyle, true);
		gcodePaneCurrentLineStyle = gcodePaneStyledDocument.addStyle("current",
				gcodePaneDefaultStyle);
		StyleConstants.setBold(gcodePaneCurrentLineStyle, true);
		gcodePaneStyledDocument.insertString(0, gcodeReader.getGCodeAsString(),
				gcodePaneDefaultStyle);

		// Position indicators
		c.gridx = 0;
		c.gridy = 2;
		c.gridheight = 1;
		c.weighty = 1.0;
		JPanel positionIndicatorsPanel = new JPanel();
		positionIndicatorsPanel.setLayout(new BoxLayout(
				positionIndicatorsPanel, BoxLayout.PAGE_AXIS));
		positionIndicatorsPanel.add(new JLabel("Absolute position:"));
		absPositionIndicator = new JLabel();
		positionIndicatorsPanel.add(absPositionIndicator, c);
		positionIndicatorsPanel.add(new JLabel("Relative to datum:"));
		relPositionIndicator = new JLabel();
		positionIndicatorsPanel.add(relPositionIndicator);
		panel.add(positionIndicatorsPanel, c);
		setPositionIndicator();

		// Status field
		c.gridy = 3;
		c.gridwidth = 2;
		c.fill = GridBagConstraints.BOTH;
		statusField = new JTextField();
		statusField.setEditable(false);
		statusField.setPreferredSize(new Dimension(300, 20));
		statusField.setMaximumSize(new Dimension(300, 20));
		panel.add(statusField, c);

		// Run, stop, reset, setup buttons
		c.fill = GridBagConstraints.NONE;
		c.weighty = 0.0;
		c.gridx = 0;
		c.gridwidth = 1;
		c.gridy = 4;
		JPanel actionButtonsPanel = new JPanel(new FlowLayout());
		runButton = new JButton("Go");
		runButton.addActionListener(this);
		actionButtonsPanel.add(runButton);
		emergencyStopButton = new JButton("STOP");
		emergencyStopButton.addActionListener(this);
		actionButtonsPanel.add(emergencyStopButton);
		resetButton = new JButton("Reset");
		resetButton.addActionListener(this);
		actionButtonsPanel.add(resetButton);
		setupButton = new JButton("Setup");
		setupButton.addActionListener(this);
		actionButtonsPanel.add(setupButton);
		panel.add(actionButtonsPanel, c);

		// G-code input field and OK button
		c.gridx = 2;
		JPanel gcodeInputPanel = new JPanel(new GridBagLayout());
		GridBagConstraints c3 = new GridBagConstraints();
		c3.anchor = GridBagConstraints.FIRST_LINE_START;
		c3.gridwidth = 2;
		gcodeInputPanel.add(new JLabel("Manual G-code entry"), c3);
		c3.gridy = 1;
		c3.gridwidth = 1;
		gcodeInputField = new JTextField(20);
		gcodeInputPanel.add(gcodeInputField, c3);
		c3.gridx = 1;
		gcodeInputOK = new JButton("OK");
		gcodeInputPanel.add(gcodeInputOK, c3);
		gcodeInputOK.addActionListener(this);
		panel.add(gcodeInputPanel, c);

		setContentPane(panel);
		
		pack();
	}

	/**
	 * GUI state changed event handler
	 */
	public void stateChanged(ChangeEvent evt) {
	}

	/**
	 * GUI action event handler
	 */
	public void actionPerformed(ActionEvent evt) {
		try {
			if (evt.getSource() == quitMenuItem ) {
				dispose();
			}
			if (evt.getSource() == aboutMenuItem ) {
				JOptionPane.showMessageDialog(frame, "G-code for RepRap version " + version + "\n" +
						"(c) Chris Meighan 2008\n" +
						"Uses code from the RepRap project [http://reprap.org/]\n" +
						"License: GPL 3.0","About G-code for RepRap",JOptionPane.INFORMATION_MESSAGE);
			}
			if (evt.getSource() == fileOpenButton) { // Open... button
				String lastDir = p.getProperty("LastDir");
				if (fc == null) {
					if (lastDir == null ) fc = new JFileChooser();
					else fc = new JFileChooser(lastDir);
					fc.setFileFilter(new GCodeFileFilter());
				}
				int returnVal = fc.showDialog(Main.this, "Open");
				
				// Store last used directory
				File dir = fc.getCurrentDirectory();
				p.setProperty("LastDir",dir.getAbsolutePath());
				p.save();

				// File selected
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					filePath.setText(fc.getSelectedFile().getAbsolutePath());
					try {
						gcodeReader
								.load(fc.getSelectedFile().getAbsolutePath());
						
						gcodeTextPane.setText("");
						
						gcodePaneStyledDocument.insertString(0, gcodeReader.getGCodeAsString(),
								gcodePaneDefaultStyle);
						
					} catch (Exception e) {
					}
				}

				// Reset the file chooser for the next time it's shown.
				fc.setSelectedFile(null);
			} else if (evt.getSource() == runButton) { // Run button
				runProgram();
			} else if (evt.getSource() == emergencyStopButton) { // Emergency
				// stop
				// button
				stopProgram();
			} else if (evt.getSource() == resetButton) { // Reset button
				resetProgram();
			} else if (evt.getSource() == setupButton) { // Setup button
				JFrame frame = new JFrame("Setup");
				new SetupDialog(frame, this);
			} else if (evt.getSource() == incX) { // +X
				absX += Xinc;
				if (commsOK)
					motorX.seek(maxSpeedX, (int) (stepsPerUnitX * absX));
				setPositionIndicator();
			} else if (evt.getSource() == decX && (absX - Xinc) >= 0) { // -X
				absX -= Xinc;
				if (commsOK)
					motorX.seek(maxSpeedX, (int) (stepsPerUnitX * absX));
				setPositionIndicator();
			} else if (evt.getSource() == incY) { // +Y
				absY += Yinc;
				if (commsOK)
					motorY.seek(maxSpeedY, (int) (stepsPerUnitY * absY));
				setPositionIndicator();
			} else if (evt.getSource() == decY && (absY - Yinc) >= 0) { // -Y
				absY -= Yinc;
				if (commsOK)
					motorY.seek(maxSpeedY, (int) (stepsPerUnitY * absY));
				setPositionIndicator();
			} else if (evt.getSource() == incZ) { // +Z
				absZ += Zinc;
				if (commsOK)
					motorZ.seek(maxSpeedZ, (int) (stepsPerUnitZ * absZ));
				setPositionIndicator();
			} else if (evt.getSource() == decZ && (absZ - Zinc) >= 0) { // -Z
				absZ -= Zinc;
				if (commsOK)
					motorZ.seek(maxSpeedZ, (int) (stepsPerUnitZ * absZ));
				setPositionIndicator();
			} else if (evt.getSource() == homeX) { // Home X axis
				if (commsOK)
					motorX.homeReset(maxSpeedX);
				absX = 0;
				setPositionIndicator();
			} else if (evt.getSource() == homeY) { // Home Y axis
				if (commsOK)
					motorY.homeReset(maxSpeedY);
				absY = 0;
				setPositionIndicator();
			} else if (evt.getSource() == homeZ) { // Home Z axis
				if (commsOK)
					motorZ.homeReset(maxSpeedZ);
				absZ = 0;
				setPositionIndicator();
			} else if (evt.getSource() == homeAll) { // Home all axes
				if (commsOK) {
					motorX.homeReset(maxSpeedX);
					motorY.homeReset(maxSpeedY);
					motorZ.homeReset(maxSpeedZ);
				}
				absX = 0;
				absY = 0;
				absZ = 0;
				setPositionIndicator();
			} else if (evt.getSource() == setDatum) { // Set datum
				datumXoff = absX;
				datumYoff = absY;
				datumZoff = absZ;
				setPositionIndicator();
			} else if (evt.getSource() == gcodeInputOK) { // OK clicked for
				// manual input
				GCodeExecutionThread executionThread = new GCodeExecutionThread(
						gcodeInputField.getText());
				programRunning = true;
				executionThread.start();
			}
		} catch (Exception e) {
		}
	}

	/**
	 * Start running the G-code
	 */
	public void runProgram() {
		GCodeExecutionThread executionThread = new GCodeExecutionThread();
		programRunning = true;
		executionThread.start();
	}

	/**
	 * Stop the G-code
	 */
	public void stopProgram() {
		logMessage("Program stopped");
		programRunning = false;
	}

	/**
	 * Reset the G-code program and rewind to beginning
	 */
	public void resetProgram() {
		gcodeReader.setPointer(0);
		programRunning = false;
		gcodePaneStyledDocument.setCharacterAttributes(0, gcodeTextPane
				.getText().length(), gcodePaneDefaultStyle, true);
		statusField.setText("");
	}

	/**
	 * Set the position indicator displays
	 */
	public void setPositionIndicator() {
		absPositionIndicator.setText(String.format("X:%5.2f Y:%5.2f Z:%5.2f",
				absX, absY, absZ));
		relPositionIndicator.setText(String.format("X:%5.2f Y:%5.2f Z:%5.2f",
				absX - datumXoff, absY - datumYoff, absZ - datumZoff));
	}

	/**
	 * Get vector angle for x and y coordinates
	 * 
	 * @return angle in radians
	 */
	public double getVectorAngle(double x, double y) {
		double angle;
		if (Math.abs(y) < 0.0001)
			angle = Math.PI / 2;
		else
			angle = Math.abs(Math.atan(x / y));

		if (y < 0)
			angle = Math.PI - angle;
		if (x < 0)
			angle = -angle;

		return angle;
	}
	
	/**
	 * Filter for G-code files
	 */
	public class GCodeFileFilter extends FileFilter {
		public String getDescription() {
			return "Filter for G-code files";
		}
		
		public boolean accept( File file ) {
			if (file.isDirectory()) return true;
			String extension = getExtension(file);
			if ( extension == null ) return false;
			return extension.equals("nc");
		}
		
	}
	
	/**
	 * Get extension of a file
	 * @param f
	 * @return
	 */
	public static String getExtension(File f) {
        String ext = null;
        String s = f.getName();
        int i = s.lastIndexOf('.');

        if (i > 0 &&  i < s.length() - 1) {
            ext = s.substring(i+1).toLowerCase();
        }
        return ext;
    }

	/**
	 * Thread to run G-code program
	 * 
	 * @author chris
	 * 
	 */
	public class GCodeExecutionThread extends Thread {
		private int mode;
		public static final int LINE_MODE = 0;
		public static final int FILE_MODE = 1;
		private String line;
		private int speed;

		/**
		 * Create thread to run a single line of G-code
		 */
		GCodeExecutionThread(String line) {
			mode = LINE_MODE;
			this.line = line;
		}

		/**
		 * Create thread to run the currently loaded G-code program
		 */
		GCodeExecutionThread() {
			mode = FILE_MODE;
		}

		/**
		 * Run the currently loaded G-code program
		 */
		public void run() {
			GCodeReader.GCodeWordBlockList wordBlocks;
			speed = (int) (defaultFeed / distancePerRevolution);
			logMessage("Program started");
			if (mode == FILE_MODE) { // Run currently loaded G-code file
				while ((wordBlocks = gcodeReader.readLine()) != null && programRunning) {
					int filePosition = gcodeReader.getFilePosition();
					int lineLength = gcodeReader.getCurrentLineText().length();
					gcodePaneStyledDocument.setCharacterAttributes(
							filePosition, lineLength,
							gcodePaneCurrentLineStyle, true);
					runGCodeWordBlocks(wordBlocks);
					if (programRunning) gcodePaneStyledDocument.setCharacterAttributes(
							filePosition, lineLength,
							gcodePaneProcessedLineStyle, true);
					// break;
				}
				if (programRunning) logMessage("Program finished");
			} else if (mode == LINE_MODE) { // Run a single line of G-code
				wordBlocks = gcodeReader.new GCodeWordBlockList(line);
				runGCodeWordBlocks(wordBlocks);
				logMessage("Ran code "+line);
			}
			programRunning = false;
		}

		/**
		 * Run a list of G-code word blocks
		 * 
		 * @param wordBlocks
		 */
		private void runGCodeWordBlocks(
				GCodeReader.GCodeWordBlockList wordBlocks) {
			int b;
			GCodeReader.GCodeWordBlock wordBlock;
			ArrayList<Coordinate3D> movementStack = new ArrayList<Coordinate3D>();
			double newX = absX, newY = absY, newZ = absZ;
			int maxSpeed = 255;
			for (b = 0; b < wordBlocks.size(); b++) {
				if (!programRunning)
					return; // Check run flag not cleared
				boolean rapid = false;
				wordBlock = wordBlocks.get(b);
				movementStack.clear();
				if (wordBlock.containsKey('G')) { // G-code
					int value = wordBlock.get('G').intValue();
					// System.out.format("G-code %d\n",value);
					newX = wordBlock.containsKey('X') ? wordBlock.get('X')
							+ datumXoff : absX;
					newY = wordBlock.containsKey('Y') ? wordBlock.get('Y')
							+ datumYoff : absY;
					newZ = wordBlock.containsKey('Z') ? wordBlock.get('Z')
							+ datumZoff : absZ;
					switch (value) {
					case 0: // G00 - Rapid feed
						logMessage(String.format(
								"G00 - Rapid feed to X:%5.2f Y:%5.2f Z:%5.2f",
								newX-datumXoff, newY-datumYoff, newZ-datumZoff));
						movementStack.add(new Coordinate3D(newX, newY, newZ));
						rapid = true;
						break;
					case 1: // G01 - Linear interpolation
						logMessage(String
								.format(
										"G01 - Linear interpolation to X:%5.2f Y:%5.2f Z:%5.2f",
										newX-datumXoff, newY-datumYoff, newZ-datumZoff));
						movementStack.add(new Coordinate3D(newX, newY, newZ));
						rapid = false;
						break;
					case 2: // G02 - Clockwise arc
					case 3: // G03 - Counterclockwise arc
						double centreX,
						centreY;
						if (wordBlock.containsKey('R')) {
							centreX = absX + wordBlock.get('R');
							centreY = absY;
							// DOESN'T WORK RIGHT NOW, SEEMS PRETTY UNCOMMON TO
							// SPECIFY RADIUS INSTEAD OF I+J ANYWAY
						} else {
							centreX = absX + wordBlock.get('I'); // Relative
							// coordinate
							centreY = absY + wordBlock.get('J'); // Relative
							// coordinate
						}
						double aX = (absX - centreX),
						aY = (absY - centreY),
						bX = (newX - centreX),
						bY = (newY - centreY);

						double angleA,
						angleB;
						if (value == 2) { // Clockwise
							angleA = getVectorAngle(aX, aY);
							angleB = getVectorAngle(bX, bY);
						} else { // Counterclockwise
							angleB = getVectorAngle(aX, aY);
							angleA = getVectorAngle(bX, bY);
						}

						if (angleB <= angleA)
							angleB += 2 * Math.PI;
						double angle = angleB - angleA;

						double radius = Math.sqrt(aX * aX + aY * aY);
						double length = radius * angle;
						int steps = (int) Math.ceil(length / curveSection);
						int s,
						step;
						for (s = 1; s <= steps; s++) {
							step = (value == 2) ? s : steps - s; // Work
							// backwards
							// for CCW
							// arc
							newX = centreX
									+ radius
									* Math.sin(angleA + angle
											* ((double) step / steps));
							newY = centreY
									+ radius
									* Math.cos(angleA + angle
											* ((double) step / steps));
							movementStack
									.add(new Coordinate3D(newX, newY, absZ));
						}
						if (value == 2) {
							logMessage(String
									.format(
											"G02 - Clockwise arc centre X:%5.2f Y:%5.2f start angle %5.2fR sweep angle %5.2fR radius %5.2f length %5.2f units in %d steps",
											centreX - datumXoff, centreY
													- datumYoff, angleA, angle,
											radius, length, steps));
						} else {
							logMessage(String
									.format(
											"G03 - Counterclockwise arc centre X:%5.2f Y:%5.2f start angle %5.2fR sweep angle %5.2fR radius %5.2f length %5.2f units in %d steps",
											centreX - datumXoff, centreY
													- datumYoff, angleA, angle,
											radius, length, steps));
						}
						break;
					}

				}
				if (wordBlock.containsKey('M')) { // M-code

				}
				if (wordBlock.containsKey('F')) { // Set feed rate
					Double value = wordBlock.get('F');
					logMessage(String.format(
							"Feed rate set to %5.2f units/minute", value));
					speed = (int) (value / distancePerRevolution);
				}
				int m;
				for (m = 0; m < movementStack.size(); m++) {
					if (!programRunning)
						return; // Check run flag not cleared
					try {
						int moveSpeed;
						if (rapid)
							moveSpeed = maxSpeed;
						else
							moveSpeed = speed;
						Coordinate3D coord = movementStack.get(m);
						newX = coord.X;
						newY = coord.Y;
						newZ = coord.Z;
						logMessage(String.format(
								"movement to X:%5.2f Y:%5.2f Z:%5.2f", newX
										- datumXoff, newY - datumYoff, newZ
										- datumZoff));

						double offX = Math.abs(newX - absX), offY = Math
								.abs(newY - absY), offZ = Math.abs(newZ - absZ);
						if (offX > 0 && offY == 0 && offZ == 0) { // X only
							// movement
							moveSpeed = Math.min(moveSpeed, maxSpeedX);
							absX = newX;
							logMessage(String.format(
									"X axis moving to point %5.2f", absX
											- datumXoff));
							if (commsOK)
								motorX.seekBlocking(maxSpeedX,
										(int) (stepsPerUnitX * absX));
						} else if (offX == 0 && offY > 0 && offZ == 0) { // Y
							// only
							// movement
							moveSpeed = Math.min(moveSpeed, maxSpeedY);
							absY = newY;
							logMessage(String.format(
									"Y axis moving to point %5.2f", absY
											- datumYoff));
							if (commsOK)
								motorY.seekBlocking(maxSpeedY,
										(int) (stepsPerUnitY * absY));
						} else if (offX == 0 && offY == 0 && offZ > 0) { // Z
							// only
							// movement
							moveSpeed = Math.min(moveSpeed, maxSpeedZ);
							absZ = newZ;
							logMessage(String.format(
									"Z axis moving to point %5.2f", absZ
											- datumZoff));
							if (commsOK)
							 motorZ.seekBlocking(maxSpeedZ,
							 (int) (stepsPerUnitZ * absZ));
						} else if (offX > 0 && offY > 0 && offZ == 0) {
							// X & Y movement, cannot currently deal
							// with movement in three axes simultaneously

							// Length of vector
							double vectorLength = Math.sqrt(offX * offX + offY
									* offY + offZ * offZ);

							// Multiplier to apply to offsets to get desired
							// speed
							double speedMultiplier = moveSpeed / vectorLength;

							// If the speed for any axis is above the maximum
							// for that axis,
							// reduce the overall speed to fit
							if (speedMultiplier * offX > maxSpeedX)
								speedMultiplier = maxSpeedX / offX;
							if (speedMultiplier * offY > maxSpeedY)
								speedMultiplier = maxSpeedY / offY;
							if (speedMultiplier * offZ > maxSpeedZ)
								speedMultiplier = maxSpeedZ / offZ;

							// Calculate individual axis speeds
							int speedX = (int) (speedMultiplier * offX);
							int speedY = (int) (speedMultiplier * offY);
							int speedZ = (int) (speedMultiplier * offZ);
							// System.out.format("Speed X:%d\n", speedX);
							// System.out.format("Speed Y:%d\n", speedY);
							// System.out.format("Speed Z:%d\n", speedZ);

							// Recalculate movement speed
							moveSpeed = (int) (speedMultiplier * vectorLength);

							// FOLLOWING CODE BASED ON LinePrinter.java IN
							// REPRAP SOURCES
							GenericStepperMotor master, slave;

							int mEnd, sStart, sEnd;
							int movementSpeed;

							if (offX > offY) {
								master = motorX;
								slave = motorY;
								mEnd = (int) (stepsPerUnitX * newX);
								sStart = (int) (stepsPerUnitY * absY);
								sEnd = (int) (stepsPerUnitY * newY);
								movementSpeed = speedX;
								/*
								 * System.out .format( "Interpolated motion with
								 * X as master, speed %d\n", speedX);
								 */
							} else {
								master = motorY;
								slave = motorX;
								mEnd = (int) (stepsPerUnitY * newY);
								sStart = (int) (stepsPerUnitX * absX);
								sEnd = (int) (stepsPerUnitX * newX);
								movementSpeed = speedY;
								/*
								 * System.out .format( "Interpolated motion with
								 * Y as master, speed %d\n", speedY);
								 */
							}

							logMessage(String.format(
									"Diagonal movement to point X:%5.2f Y:%5.2f Z:%5.2f", absX
									- datumXoff, absY
									- datumYoff, absZ
											- datumZoff));
							
							if (commsOK) {
								master.setSync(GenericStepperMotor.SYNC_NONE);
								if (sStart < sEnd)
									slave.setSync(GenericStepperMotor.SYNC_INC);
								else
									slave.setSync(GenericStepperMotor.SYNC_DEC);

								int deltaY = Math.abs(sEnd - sStart);
								master.dda(movementSpeed, mEnd, deltaY);
								slave.setSync(GenericStepperMotor.SYNC_NONE);
							}

							absX = newX;
							absY = newY;
							absZ = newZ;
						} else if ( offX == 0 && offY == 0 && offZ == 0 ) {
							logMessage("No movement");
						} else {
							logMessage("Error - unsupported movement");
						}
						//logMessage(String.format(
						//		"Moving to X%5.2f Y%5.2f Z%5.2f at speed %d",
						//		absX - datumXoff, absY - datumYoff, absZ
						//				- datumZoff, moveSpeed));
						setPositionIndicator();
					} catch (Exception e) {
						logMessage(String.format("Error:%s", e.getMessage()));
					}
				}
			}
		}
	}

	/**
	 * Three dimensional coordinate
	 * 
	 * @author chris
	 * 
	 */
	private class Coordinate3D {
		Double X, Y, Z;

		public Coordinate3D(double X, double Y, double Z) {
			this.X = X;
			this.Y = Y;
			this.Z = Z;
		}
	}

	/**
	 * Load preferences
	 */
	public void loadPreferences() throws Exception {

		p.load();
		
		// Port name
		portName = p.getProperty("Port(name)");

		// Addresses
		XAxisAddress = Integer.valueOf(p.getProperty("XAxisAddress"));
		YAxisAddress = Integer.valueOf(p.getProperty("YAxisAddress"));
		ZAxisAddress = Integer.valueOf(p.getProperty("ZAxisAddress"));

		// Jog increment
		Xinc = Double.valueOf(p.getProperty("JogIncrement"));
		Yinc = Double.valueOf(p.getProperty("JogIncrement"));
		Zinc = Double.valueOf(p.getProperty("JogIncrement"));

		// Steps per unit
		stepsPerUnitX = (int) (Double.valueOf(p.getProperty("XAxisSteps")) / Double
				.valueOf(p.getProperty("DistancePerRev")));
		stepsPerUnitY = (int) (Double.valueOf(p.getProperty("YAxisSteps")) / Double
				.valueOf(p.getProperty("DistancePerRev")));
		stepsPerUnitZ = (int) (Double.valueOf(p.getProperty("ZAxisSteps")) / Double
				.valueOf(p.getProperty("DistancePerRev")));

		// Maximum motor speed
		maxSpeedX = Integer.valueOf(p.getProperty("XAxisMaxSpeed"));
		maxSpeedY = Integer.valueOf(p.getProperty("YAxisMaxSpeed"));
		maxSpeedZ = Integer.valueOf(p.getProperty("ZAxisMaxSpeed"));

		// Default feed rate (units/minute)
		defaultFeed = Double.valueOf(p.getProperty("DefaultFeedRate"));

		// Curve section length (units)
		curveSection = Double.valueOf(p.getProperty("CurveSectionLength"));
		;

		// Distance travelled per revolution of motor (units/revolution)
		distancePerRevolution = Double.valueOf(p.getProperty("DistancePerRev"));

		// Reinitialise comms and motors, since the port and/or addresses may
		// have been changed
		if (motorX != null)
			motorX.dispose();
		if (motorY != null)
			motorY.dispose();
		if (motorZ != null)
			motorZ.dispose();
		if (communicator != null)
			communicator.dispose();
		initialiseComms();
		initialiseMotors();
	}

	/**
	 * Log a message
	 * 
	 * @param message
	 */
	private void logMessage(String message) {
		System.out.println(message);
		statusField.setText(message);
	}
}